; Tribes2 Query mIRC script v1.0.0 by Jason "VeKToR" Gill.
; April 10, 2001
; Copyright (C) 2001 Jason Gill

; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.

; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.

; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
; http://www.gnu.org/copyleft/gpl.html

; ----------------------------------------------------------------------------
; Documentation
; ----------------------------------------------------------------------------

; -------------------------
; /T2_QueryServer <ip> <port> <alias> <timeoutalias>
; Queries a SINGLE server and stores the resulting info in the
; %T2-Info.* variables. Triggers the alias <alias> when complete.
; If the request times out, <timeoutalias> is called
; -------------------------

; -------------------------
; $T2_Player(<player-number>, <field>)
; where field is any of name, flags, score, team.
; Returns information about a given player without having to screw around with
; mIRC's icky messy variable formats. 
; The following fields return the following information:
; name: the player's name, a string
; flags: a space-delimited string containing: registered, alias, bot
; score: the player's score, a number
; team: the player's team name preceded by the player's team number.
; -------------------------

; -------------------------
; $T2_Team(<team-number>, <field>)
; where field is any of name, size, score
; Returns information about a given team without having to screw around with
; mIRC's icky messy variable formats. 
; The following fields return the following information:
; name: the player's name, a string
; size: the number of players on the team, a number
; score: the player's score, a number
; -------------------------

; -------------------------
; The variables:
; when the T2_QueryServer calls your defined alias (the callback alias, or 
; the callback), it fills a series of variables with information regarding the
; server's state that you may use. The following is a list of these variables:
;
; %T2-Info.Ping: An estimate of the response time of the server, in ms.
; %T2-Info.Response: the response type sent back by the server. Should be 20.
; %T2-Info.Response2: the response type for the second packet. Should be 16.
; %T2-Info.Flags: The query flags sent by the server. Should be 2.
; %T2-Info.Flags2: The query flags for the second packet. Should be 2.
; %T2-Info.QID: The Query-ID echoed back by the server. Arbitrary 4 bytes. Should be 1 1 1 1
; %T2-Info.QID2: The Query-ID for the second packet. Arbitrary 4 bytes. Should be 1 1 1 1
; %T2-Info.mod: The mod the server is currently running. Example: base
; %T2-Info.mapType: The map type the server is running. Examples: Rabbit, Siege
; %T2-Info.map: The actual map the server is running. Examples: Firestorm, Riverdance
; %T2-Info.statusBits: The raw status bits of the server. More below.
; %T2-Info.status: Interpreted status bits. A space-delimited string containing any of the
;                     following words: dedicated password tournament no-alias
; %T2-Info.players: The number of players on the server.
; %T2-Info.maxPlayers: The server's player limit. Maximum number of players allowed.
; %T2-Info.bots: The number of bots on the server.
; %T2-Info.CPU: The cpu speed of the server.
; %T2-Info.infoString: A short blurb set by the server's admin. Can contain just about anything.
; %T2-Info.numTeams: The number of teams on the server (not counting unassigned)
; %T2-Info.queryVersion: The query protocol version. Should be VER3. If it's not, email me at vektor@linux.ca
;                           so I can see about updating this query script.
; %T2-Info.netProtocol: Unsure of this one, but I think it's the Tribes2 network code version used to
;                          determine incompatable clients and servers.
; %T2-Info.version: The version (build number) of the server.
; %T2-Info.name: The name of the server (currently appears to be broken. I'm looking into it)
;
; [RS]
; %T2-Info.validServerInfo: Says whether the server information is valid.
; ALWAYS MAKE SURE THIS IS 1 BEFORE GETTING TEAM/PLAYER INFORMATION!
; 
; The rest of the variables require ugly formatting and are accessibly by two convenience
; aliases $T2_Player and $T2_Team
; -------------------------

; -------------------------
; Now, about those status bits:
; If bit 1 is set, the server is dedicated.
; If bit 2 is set, the server is passworded.
; If bit 7 is set, the server is in tournament mode
; If bit 8 is set, the server doesn't allow aliases.
; -------------------------

; ----------------------------------------------------------------------------
; Interface Section
; ----------------------------------------------------------------------------

; ----------
; The alias that actually starts the server query. This is the first thing a client should call.
; ----------
alias T2_QueryServer {
  if ($sock(T2Query).name != $null) sockclose T2Query
  if ($sock(T2Query2).name != $null) sockclose T2Query2

  ; Since we have to keep track of both queries, and wait 'till both are done.
  set %T2Query.stage 0

  ; The proper bytes to instruct a server to return info to us!
  bset &QueryPacket 1 18 2 1 1 1 1
  bset &QueryPacket2 1 14 2 1 1 1 1

  ; Set up some values we'll need to know later: 
  ; the time the query was sent, and what alias to call when we're done!
  set %T2Query.PingInit $ticks
  set %T2Query.QueryParseAlias $3
  unset %T2-Info.*

  ; Actually send our query to the server (ip:port) indicated by the first 2 params
  sockudp -kb T2Query $1 $2 6 &QueryPacket
  sockudp -kb T2Query2 $1 $2 6 &QueryPacket2

  ; Setup a 8-second timer that triggers to tell us our query timed out.
  .timerT2SingleParse 1 8 T2_QueryTimeout $4 $1 $2
}

; ----------
; Retrieves information about a player - you could do it with variable names, but mIRC has horrid syntax
; for handling that sort of thing.
; Takes 2 params, first is the player number, second is the info field required
; (name, flags, team, score)
; ----------
alias T2_Player {
  return [ %T2-Info.Player- [ $+ [ $2 ] [ $+ [ $1 ] ] ] ]
}

; ----------
; Retrieves information about a team - you could do it with variable names, but mIRC has horrid syntax
; for handling that sort of thing.
; Takes 2 params, first is the team number, second is the info field required
; (name, size, score)
; ----------
alias T2_Team {
  return [ %T2-Info.Team- [ $+ [ $2 ] [ $+ [ $1 ] ] ] ]
}

; ----------------------------------------------------------------------------
; Internal Section
; ----------------------------------------------------------------------------

; ----------
; The alias that handles query timeouts. Calls the client's specified timeout-callback alias
; ----------
alias T2_QueryTimeout {
  sockclose T2SingleQuery
  if ($isalias($1) == $true) $1 $2 $3
}

; ----------
; Reads a Pascal-format string starting at the byte specified by it's first param
; or %T2Query.byte (position in the binvar) if none specified
; ----------
alias -l ReadPascalString {
  if ($1 != $null) {
    %T2Query.byte = $calc($1 + $bvar(&packet, $1) + 1)
    var T2Query.range $calc($1 + 1) $+ - $+ $calc(%T2Query.byte - 1)
  }
  else {
    var %old %T2Query.byte
    %T2Query.byte = $calc(%T2Query.byte + $bvar(&packet, %T2Query.byte) + 1)
    var %T2Query.range $calc(%old + 1) $+ - $+ $calc(%T2Query.byte - 1)
  }
  return $bvar(&packet, %T2Query.range).text
}

; ----------
; Reads a number of bytes specified by the second param starting at a pos specified by the first
; from the binvar
; If only one param specified, it reads that number of bytes starting at %T2Query.byte 
; If the last param is "text" (omitted from the param# check) then it returns the ASCII text rep of the bin
; ----------
alias -l ReadBinary {
  if ($gettok($1-, $numtok($1-, 32), 32) == text) {
    var %return_text $true
    var %checkparm $3
  }
  else var %checkparm $2

  if (%checkparm != $null) {
    %T2Query.byte = $calc($1 + $2)
    var %T2Query.range $1 $+ - $+ $calc(%T2Query.byte - 1)
  }
  else {
    var %T2Query.range %T2Query.byte $+ - $+ $calc(%T2Query.byte + $1 - 1)
    inc %T2Query.byte $1
  }

  if (%return_text == $null) return $bvar(&packet, %T2Query.range)
  else return $bvar(&packet, %T2Query.range).text
}

; ----------
; Seeks forward in the binvar until it finds a value matching the 1st param, and sets Lastbyte to its position
; if the second param is "y"
; returns $null (and does not set lastbyte) if no char found
; ----------
alias -l SeekBin {
  var %pos = $bfind(&packet, %T2Query.byte, $1) 

  if (%pos != 0) {
    if ($2 == y) %T2Query.byte = %pos
    else return %pos
  }
  else {
    return $null
  }
}

; ----------
; This is the alias that actually does the packet parsing.
; It's quite ugly.
; ----------
alias T2_ParseQueryPacket {  
  ; Server response byte - it should be 20
  %T2-Info.Response = $ReadBinary(1, 1)

  ; Server info flags - it should be 2, meaning uncompressed strings
  %T2-Info.Flags = $ReadBinary(1)

  ; Server Query ID - it should be 4 bytes (1 1 1 1) - actually, a 32-bit int.
  ; This is an arbitrary value I just plugged in
  %T2-Info.QID = $ReadBinary(4)

  ; This is the mod the server is running.
  %T2-Info.Mod = $ReadPascalString

  ; This is the map type that's currently playing on the server (Rabbit, Siege, etc)
  %T2-Info.MapType = $ReadPascalString

  ; This is the actual map that's running on the server (Firestorm, Riverdance, etc)
  %T2-Info.Map = $ReadPascalString

  ; This is a status bitmask that tells us several things about the server: see below
  %T2-Info.StatusBits = $readBinary(1)

  if (1 & %T2-Info.StatusBits) set %T2-Info.Status dedicated
  if (2 & %T2-Info.StatusBits) set %T2-Info.Status %T2-Info.Status password
  if (64 & %T2-Info.StatusBits) set %T2-Info.Status %T2-Info.Status tournament
  if (128 & %T2-Info.StatusBits) set %T2-Info.Status %T2-Info.Status no-alias

  ; Number of players on the server
  %T2-Info.Players = $readBinary(1)

  ; Maximum number of players on the server
  %T2-Info.MaxPlayers = $readBinary(1)

  ; Number of bots on the server
  %T2-Info.Bots = $readBinary(1)

  ; 16-bit unsigned int for cpu speed
  %T2-Info.CPU = $calc($readBinary(1) + $calc($readBinary(1)*256))

  ; Server's info string. It's just a message set by the server's admin.
  %T2-Info.InfoString = $readPascalString

  ; Length in bytes of the "content" field of the server reply. I currently don't use it,
  ; but it can't hurt to keep it here anyway in case I need it later.
  var %servercontent = $calc($readBinary(1) + $calc($readBinary(1)*256))

  ; The number of teams on the server. It'll usually be 2, but we gotta check anyway!
  %T2-Info.NumTeams = $ReadBinary(1, text)

  ; [RS] A hack to stop it from breaking
  if ( %T2-Info.NumTeams isnum ) {
    ; TODO: Make convenience functions for reading team info
    ; This part reads the names and score(?) info for all the teams and stores them.
    var %loop = 1
    while (%loop <= %T2-Info.NumTeams) {
      $Seekbin(10,y)
      inc %T2Query.byte
      var %len = $calc($Seekbin(9) - %T2Query.byte)
      set %T2-Info.Team-name $+ %loop $ReadBinary(%len, text)

      ; [RS] This block is what has been changed.
      ; inc %T2Query.byte
      ; set %T2-Info.Team-score $+ %loop $readBinary(1, text)
      $Seekbin(9,y)
      inc %T2Query.byte
      var %len = $calc($Seekbin(10) - %T2Query.byte)
      set %T2-Info.Team-score $+ %loop $readBinary(%len, text)

      inc %loop 1
    }

    ; Although we already got the #players on the server, this byte tells us what it is again
    ; Since this is packaged with the actual player data, I trust it more.
    $Seekbin(10,y)
    inc %T2Query.byte
    var %len = $calc($Seekbin(10) - %T2Query.byte)
    var %entries = $ReadBinary(%len,text)


    ; [RS] ANOTHER hack to stop it from breaking
    if ( %entries isnum ) {
      ; TODO: Make convenience functions for reading player info
      ; This loops through the raw player info and extracts name, team and score.
      ; From the name, it determines if it's a bot, registered name or alias.
      ; From the team name, it determines what team number the player is on and the # players on that team
      ; (just inc by 1 for each player on a given team)
      var %loop = 1

      while (%loop <= %entries) {
        $Seekbin(10,y)
        inc %T2Query.byte 2
        var %len = $calc($Seekbin(9) - %T2Query.byte)

        ; Using a temp variable to do this because of mIRC's ungodly syntax like
        ; [ %T2-Info.player-name [ $+ [ %loop ] ] ]
        var %tmp = $ReadBinary(%len,text)
        set %T2-Info.player-flags $+ %loop $iif($chr(8) isin %tmp, registered, $chr(32)) $iif($chr(12) isin %tmp, alias, $chr(32)) $iif($chr(14) isin %tmp, bot, $chr(32))
        %tmp = $remove(%tmp, $chr(11), $chr(8), $chr(12), $chr(14), $chr(17), $chr(16))
        set %T2-Info.player-Name $+ %loop %tmp

        $Seekbin(9,y)
        inc %T2Query.byte
        var %len = $calc($Seekbin(9) - %T2Query.byte)

        ; Again with the temp variables because mIRC HAS SOME UGLY SYNTAX!@#
        %tmp = $ReadBinary(%len,text)

        set %T2-Info.player-team $+ %loop 0 none
        var %loop2 = 1
        while (%loop2 <= %T2-Info.NumTeams) {
          if ( [ %T2-Info.Team-name [ $+ [ %loop2 ] ] ] == %tmp ) {
            inc %T2-Info.Team-size $+ %loop2
            set %T2-Info.player-team $+ %loop %loop2 %tmp
            break
          }
          inc %loop2
        }

        $Seekbin(9,y)
        inc %T2Query.byte
        var %len = $calc($Seekbin(10) - %T2Query.byte)
        if (%len > 0) {
          set %T2-Info.player-score $+ %loop $ReadBinary(%len,text)
        }
        else {
          ; If we can't find a 10 after our position, we must be at the and of the binvar
          ; so just read what's left as score.
          var %T2Query.range = %T2Query.byte $+ -
          set %T2-Info.player-score $+ %loop $bvar(&packet, %T2Query.range).text
        }

        inc %loop
      }
      ; [RS] A new variable for your use.
      ; 0 = server is cycling maps (or a much bigger problem...)
      set %T2-Info.validServerInfo 1
    }
    else {
      set %T2-Info.validServerInfo 0
    }

  }
  else {
    set %T2-Info.validServerInfo 0
  }

}

alias T2_ParseQueryPacket2 {  
  %T2-Info.Response2 = $ReadBinary(1,1)
  %T2-Info.flags2 = $ReadBinary(1)
  %T2-Info.QID2 = $ReadBinary(4)

  ; The Query protocol version string, this was VER3 at the time of writing.
  ; If it's ever anything else, send me an email at vektor@linux.ca so I can update or add fields as
  ; necessary.
  %T2-Info.queryVersion = $ReadPascalString

  ; Not sure what this is. Possibly a Tribes network protocol version?
  ; So we can filter between old servers with different netcode and newer patched ones?
  %T2-Info.NetProtocol = $calc($readBinary(1) + $calc($readBinary(1)*256) + $calc($readBinary(1)*65536) + $calc($readBinary(1)*16777216))

  ; The server build number (version)
  %T2-Info.version = $calc($readBinary(1) + $calc($readBinary(1)*256) + $calc($readBinary(1)*65536) + $calc($readBinary(1)*16777216))
  %T2-Info.name = $readpascalstring

}

; ----------------------------------------------------------------------------
; Net events handler
; ----------------------------------------------------------------------------

on 1:UDPREAD:T2Query:{ 
  if (%T2Query.stage > 0) .timerT2SingleParse off

  if (%T2Query.PingInit != $null) {
    %T2-Info.Ping = $calc($ticks - %T2Query.PingInit)
    unset %T2Query.PingInit
  }

  set %T2Query.ip $sock($sockname).saddr
  set %T2Query.port $sock($sockname).sport

  sockread &packet
  while ($sockbr > 0) {
    sockread -f &tmp
    if ($sockbr > 0) {
      var %position = $calc($bvar(&packet,0) + 1)
      bcopy &packet %position &tmp 1 -1
    }
  }
  sockclose $sockname

  T2_parseQueryPacket %T2Query.ip $+ : $+ %T2Query.port

  if (%T2Query.stage > 0) T2Query_UDP_Complete
  inc %T2Query.stage
}

; The other packet
on 1:UDPREAD:T2Query2:{ 
  if (%T2Query.stage > 0) .timerT2SingleParse off

  if (%T2Query.PingInit != $null) {
    %T2-Info.Ping = $calc($ticks - %T2Query.PingInit)
    unset %T2Query.PingInit
  }

  set %T2Query.ip $sock($sockname).saddr
  set %T2Query.port $sock($sockname).sport

  sockread &packet
  while ($sockbr > 0) {
    sockread -f &tmp
    if ($sockbr > 0) {
      var %position = $calc($bvar(&packet,0) + 1)
      bcopy &packet %position &tmp 1 -1
    }
  }
  sockclose $sockname

  T2_parseQueryPacket2 %T2Query.ip $+ : $+ %T2Query.port

  if (%T2Query.stage > 0) T2Query_UDP_Complete
  inc %T2Query.stage
}

alias T2Query_UDP_Complete {
  .timerT2SingleParse off

  if ($isalias(%T2Query.QueryParseAlias) == $true) %T2Query.QueryParseAlias %T2Query.ip %T2Query.port

  unset %T2-Info.*
  unset %T2Query.*
}
